#!/usr/bin/env python

"""
Confluence Wiki syntax parsing.

Copyright (C) 2012 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file LICENCE.txt
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

--------

The basic procedure is as follows:

 1. Wiki pages are first split up into regions.
 2. Then, within these regions, the text is split into blocks.
    1. First, lists are identified.
    2. Additionally, other block-like elements are identified.
 3. Each block is then parsed.
"""

import re

# Section extraction.

sections_regexp_str = r"(?<!{)(?P<section>{(?P<type>[^{}\n]+)}.*?{(?P=type)})"
sections_regexp = re.compile(sections_regexp_str, re.DOTALL | re.MULTILINE)

def get_regions(s):

    """
    Return a list of regions from 's'. Each region is specified using a tuple of
    the form (type, text).
    """

    last = 0
    regions = []
    for match in sections_regexp.finditer(s):
        start, end = match.span()
        regions.append((None, s[last:start]))
        regions.append(get_section_details(s[start:end]))
        last = end
    regions.append((None, s[last:]))
    return regions

# Section inspection.

section_regexp_str = r"{(?P<sectiontype>[^\n]*?)}(?P<section>.*){(?P=sectiontype)}"
section_regexp = re.compile(section_regexp_str, re.DOTALL | re.MULTILINE)

def get_section_details(s):

    "Return the details of a section 's' in the form (type, text)."

    match = section_regexp.match(s)
    if match:
        return match.group("sectiontype"), match.group("section")
    else:
        return None, s

# List extraction.

list_regexp_str = r"^(?P<listtype>[*#-])[*#-]*.*\n((?P=listtype).*(?:\n|$))*"
list_regexp = re.compile(list_regexp_str, re.MULTILINE)

def get_lists(s):

    """
    Extract lists from the given string 's'.
    """

    last = 0
    blocks = []
    for match in list_regexp.finditer(s):
        start, end = match.span()
        blocks.append((None, s[last:start]))
        blocks.append(("list", s[start:end]))
        last = end
    blocks.append((None, s[last:]))
    return blocks

# Block extraction.

block_regexp_str = r"^(?:\s*\n)+"
block_regexp = re.compile(block_regexp_str, re.MULTILINE)

def get_basic_blocks(s):

    """
    Return blocks from the given string 's' by splitting the text on blank lines
    and eliminating those lines.
    """

    return [b for b in block_regexp.split(s) if b.strip()]

# Block inspection.

blocktext_regexp_str = r"^(?P<type>h\d|bq)\.\s+(?P<text>.*)$"
blocktext_regexp = re.compile(blocktext_regexp_str, re.MULTILINE)

def get_blocks(s):

    """
    Return blocks from the given string 's', inspecting the basic blocks and
    generating additional block-level text where appropriate.
    """

    blocks = []

    for blocktype, blocktext in get_lists(s):

        # Collect list blocks.

        if blocktype is not None:
            blocks.append((blocktype, blocktext))

        # Attempt to find new subblocks in other regions.

        else:
            for block in get_basic_blocks(blocktext):
                last = 0
                for match in blocktext_regexp.finditer(block):
                    start, end = match.span()

                    # Add preceding non-block text.

                    preceding = block[last:start]
                    if preceding.strip():
                        blocks.append((None, preceding))

                    # Add the subblock.

                    blocks.append((match.group("type"), match.group("text")))
                    last = end

                # Add trailing non-block text.

                trailing = block[last:]
                if trailing.strip():
                    blocks.append((None, trailing))

    return blocks

listitem_regexp_str = r"^([*#-])+\s*(.*)$"
listitem_regexp = re.compile(listitem_regexp_str, re.MULTILINE)

if __name__ == "__main__":
    import sys

    s = sys.stdin.read()

    for type, text in get_regions(s):
        if type is None:
            for blocktype, blocktext in get_blocks(text):
                print "Block type:", blocktype
                print blocktext
                print
        else:
            print "Region type:", type
            print text
            print

        print "-" * 60

# vim: tabstop=4 expandtab shiftwidth=4
